/*-------------------------------------------------------------------------
 *
 * nodeSeqscan.c
 *	  Support routines for sequential scans of relations.
 *
 * Portions Copyright (c) 1996-2021, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 *
 * IDENTIFICATION
 *	  src/backend/executor/nodeSeqscan.c
 *
 *-------------------------------------------------------------------------
 */
/*
 * INTERFACE ROUTINES
 *		ExecSeqScan				sequentially scans a relation.
 *		ExecSeqNext				retrieve next tuple in sequential order.
 *		ExecInitSeqScan			creates and initializes a seqscan node.
 *		ExecEndSeqScan			releases any storage allocated.
 *		ExecReScanSeqScan		rescans the relation
 *
 *		ExecSeqScanEstimate		estimates DSM space needed for parallel scan
 *		ExecSeqScanInitializeDSM initialize DSM for parallel scan
 *		ExecSeqScanReInitializeDSM reinitialize DSM for fresh parallel scan
 *		ExecSeqScanInitializeWorker attach to DSM info in parallel worker
 */
#include "postgres.h"

#include "access/relscan.h"
#include "access/tableam.h"
#include "executor/execdebug.h"
#include "executor/nodeSeqscan.h"
#include "nodes/nodes.h"
#include "utils/rel.h"
/* CS631 start */
/* Including SPI for executing the insert statements into AIDX_QUERIES table */
#include "executor/spi.h"
/* Including for CMDType struct use */
#include "executor/execExpr.h"
/* CS631 end */

static TupleTableSlot *SeqNext(SeqScanState *node);


/* ----------------------------------------------------------------
 *						Scan Support
 * ----------------------------------------------------------------
 */

/* ----------------------------------------------------------------
 *		SeqNext
 *
 *		This is a workhorse for ExecSeqScan
 * ----------------------------------------------------------------
 */
static TupleTableSlot *
SeqNext(SeqScanState *node)
{
	TableScanDesc scandesc;
	EState	   *estate;
	ScanDirection direction;
	TupleTableSlot *slot;

	/*
	 * get information from the estate and scan state
	 */
	scandesc = node->ss.ss_currentScanDesc;
	estate = node->ss.ps.state;
	direction = estate->es_direction;
	slot = node->ss.ss_ScanTupleSlot;

	if (scandesc == NULL)
	{
		/*
		 * We reach here if the scan is not parallel, or if we're serially
		 * executing a scan that was planned to be parallel.
		 */
		scandesc = table_beginscan(node->ss.ss_currentRelation,
								   estate->es_snapshot,
								   0, NULL);
		node->ss.ss_currentScanDesc = scandesc;
	}

	/*
	 * get the next tuple from the table
	 */
	if (table_scan_getnextslot(scandesc, direction, slot))
		return slot;
	return NULL;
}

/*
 * SeqRecheck -- access method routine to recheck a tuple in EvalPlanQual
 */
static bool
SeqRecheck(SeqScanState *node, TupleTableSlot *slot)
{
	/*
	 * Note that unlike IndexScan, SeqScan never use keys in heap_beginscan
	 * (and this is very bad) - so, here we do not check are keys ok or not.
	 */
	return true;
}

/* ----------------------------------------------------------------
 *		ExecSeqScan(node)
 *
 *		Scans the relation sequentially and returns the next qualifying
 *		tuple.
 *		We call the ExecScan() routine and pass it the appropriate
 *		access method functions.
 * ----------------------------------------------------------------
 */
static TupleTableSlot *
ExecSeqScan(PlanState *pstate)
{
	SeqScanState *node = castNode(SeqScanState, pstate);

	return ExecScan(&node->ss,
					(ExecScanAccessMtd) SeqNext,
					(ExecScanRecheckMtd) SeqRecheck);
}


/* ----------------------------------------------------------------
 *		ExecInitSeqScan
 * ----------------------------------------------------------------
 */
SeqScanState *
ExecInitSeqScan(SeqScan *node, EState *estate, int eflags)
{
	SeqScanState *scanstate;

	/*
	 * Once upon a time it was possible to have an outerPlan of a SeqScan, but
	 * not any more.
	 */
	Assert(outerPlan(node) == NULL);
	Assert(innerPlan(node) == NULL);

	/*
	 * create state structure
	 */
	scanstate = makeNode(SeqScanState);
	scanstate->ss.ps.plan = (Plan *) node;
	scanstate->ss.ps.state = estate;
	scanstate->ss.ps.ExecProcNode = ExecSeqScan;


	/*
	 * Miscellaneous initialization
	 *
	 * create expression context for node
	 */
	ExecAssignExprContext(estate, &scanstate->ss.ps);

	/*
	 * open the scan relation
	 */
	scanstate->ss.ss_currentRelation =
		ExecOpenScanRelation(estate,
							 node->scanrelid,
							 eflags);

/* CS631 start */
	/*
	 * Eg: SELECT * FROM PG_DATABASE
	 * Relation name => scanstate->ss.ss_currentRelation->rd_rel->relname.data === PG_DATABASE
	 */
	char* relName = "";

	// Relation Name
	relName = scanstate->ss.ss_currentRelation->rd_rel->relname.data;

	elog(LOG, "%s", relName);
/* CS631 end */

	/* and create slot with the appropriate rowtype */
	ExecInitScanTupleSlot(estate, &scanstate->ss,
						  RelationGetDescr(scanstate->ss.ss_currentRelation),
						  table_slot_callbacks(scanstate->ss.ss_currentRelation));

/* CS631 start */
	/*
	 * List of Attributes selected => scanstate->ss.ss_ScanTupleSlot->tts_tupleDescriptor->attrs
	 */
	TupleDescData *all_selected_attributes = scanstate->ss.ss_ScanTupleSlot->tts_tupleDescriptor;

	elog(LOG, "%d", all_selected_attributes->natts);

	FormData_pg_attribute *attrs = all_selected_attributes->attrs;

	char attr_selected[1000] = "";
	int j, l, final_string_length, k, ret;
	char final_attrs[1000] = "";

	char *comma = ",", *eos="\0";
	for (int i=0; i<all_selected_attributes->natts; i++){
		j = strlen(attr_selected);
		l = 0;
		final_string_length = strlen(final_attrs);
		for(k=strlen(final_attrs); k<strlen(attrs[i].attname.data)+final_string_length; k++){
			final_attrs[k] = attrs[i].attname.data[l++];
		}
		final_attrs[k] = comma[0];
	}
	final_attrs[k] = eos[0];
	elog(LOG, "%s", final_attrs);
/* CS631 end */


	/*
	 * Initialize result type and projection.
	 */
	ExecInitResultTypeTL(&scanstate->ss.ps);
	ExecAssignScanProjectionInfo(&scanstate->ss);

	/*
	 * initialize child expressions
	 */
	scanstate->ss.ps.qual =
		ExecInitQual(node->plan.qual, (PlanState *) scanstate);
/* CS631 start */
	/*
	 * Creating the INSERT query for inserting data into AIDX_QUERIES table
	 * AIDX_QUERIES(
	 * 		type varchar(100), -- Type of the query - select/delete/insert
	 * 		tablename varchar(100), -- Table name
	 * 		attrs varchar(100) -- attributes in the where clause (currently for insert we insert * in the table))
	 * 
	 *
	 */

	char new_new_query[200] = "INSERT INTO AIDX_QUERIES VALUES('";
	char *type="";
	char *add1 = "','";
	char *add2 = "','";
	char *add3 = "');";
	char qual_attr[100] = "";
	if(scanstate->ss.ps.qual!= NULL){
		int attnum = scanstate->ss.ps.qual->steps->d.var.attnum;

	elog(LOG, "%d", attnum);


	for(int i=0; i<strlen(attrs[attnum-1].attname.data); i++){
		qual_attr[i] = attrs[attnum-1].attname.data[i];
	}
	elog(LOG, "%s", qual_attr);

	/*Retrieving the Type of the query*/
	if(estate->es_plannedstmt->commandType == CMD_SELECT){
		type =  "select";
		elog(LOG,"CMDTYPE SELECT \n");
	}
	if(estate->es_plannedstmt->commandType == CMD_DELETE){
		type =  "delete";
		elog(LOG,"CMDTYPE DELTE \n");
	}
	}
	else if(scanstate->ss.ps.qual == NULL && estate->es_plannedstmt->commandType == CMD_DELETE){
		qual_attr[0] = '*';
		type="delete";

	}

	int i1;

	int j1 = strlen(new_new_query);

	for (i1 = 0; type[i1] != '\0'; i1++) {
		new_new_query[i1 + j1] = type[i1];
	}

	j1 = strlen(new_new_query);

	for (i1 = 0; add1[i1] != '\0'; i1++) {
			new_new_query[i1 + j1] = add1[i1];
	}

	j1 = strlen(new_new_query);

	for (i1 = 0; relName[i1] != '\0'; i1++) {
			new_new_query[i1 + j1] = relName[i1];
	}

	j1 = strlen(new_new_query);

	for (i1 = 0; add2[i1] != '\0'; i1++) {
			new_new_query[i1 + j1] = add2[i1];
	}

	j1 = strlen(new_new_query);

	for (i1 = 0; qual_attr[i1] != '\0'; i1++) {
			new_new_query[i1 + j1] = qual_attr[i1];
	}

	j1 = strlen(new_new_query);

	for (i1 = 0; add3[i1] != '\0'; i1++) {
			new_new_query[i1 + j1] = add3[i1];
	}

	new_new_query[i1 + j1] = '\0';

	elog(LOG, "%s", new_new_query);

	if(type == "select" || type == "delete")
		if( SPI_connect() == SPI_OK_CONNECT )
		{
			ret = SPI_execute(new_new_query,false,0);
			ret = SPI_finish();
		}
/* CS631 end */

	return scanstate;
}

/* ----------------------------------------------------------------
 *		ExecEndSeqScan
 *
 *		frees any storage allocated through C routines.
 * ----------------------------------------------------------------
 */
void
ExecEndSeqScan(SeqScanState *node)
{
	TableScanDesc scanDesc;

	/*
	 * get information from node
	 */
	scanDesc = node->ss.ss_currentScanDesc;

	/*
	 * Free the exprcontext
	 */
	ExecFreeExprContext(&node->ss.ps);

	/*
	 * clean out the tuple table
	 */
	if (node->ss.ps.ps_ResultTupleSlot)
		ExecClearTuple(node->ss.ps.ps_ResultTupleSlot);
	ExecClearTuple(node->ss.ss_ScanTupleSlot);

	/*
	 * close heap scan
	 */
	if (scanDesc != NULL)
		table_endscan(scanDesc);
}

/* ----------------------------------------------------------------
 *						Join Support
 * ----------------------------------------------------------------
 */

/* ----------------------------------------------------------------
 *		ExecReScanSeqScan
 *
 *		Rescans the relation.
 * ----------------------------------------------------------------
 */
void
ExecReScanSeqScan(SeqScanState *node)
{
	TableScanDesc scan;

	scan = node->ss.ss_currentScanDesc;

	if (scan != NULL)
		table_rescan(scan,		/* scan desc */
					 NULL);		/* new scan keys */

	ExecScanReScan((ScanState *) node);
}

/* ----------------------------------------------------------------
 *						Parallel Scan Support
 * ----------------------------------------------------------------
 */

/* ----------------------------------------------------------------
 *		ExecSeqScanEstimate
 *
 *		Compute the amount of space we'll need in the parallel
 *		query DSM, and inform pcxt->estimator about our needs.
 * ----------------------------------------------------------------
 */
void
ExecSeqScanEstimate(SeqScanState *node,
					ParallelContext *pcxt)
{
	EState	   *estate = node->ss.ps.state;

	node->pscan_len = table_parallelscan_estimate(node->ss.ss_currentRelation,
												  estate->es_snapshot);
	shm_toc_estimate_chunk(&pcxt->estimator, node->pscan_len);
	shm_toc_estimate_keys(&pcxt->estimator, 1);
}

/* ----------------------------------------------------------------
 *		ExecSeqScanInitializeDSM
 *
 *		Set up a parallel heap scan descriptor.
 * ----------------------------------------------------------------
 */
void
ExecSeqScanInitializeDSM(SeqScanState *node,
						 ParallelContext *pcxt)
{
	EState	   *estate = node->ss.ps.state;
	ParallelTableScanDesc pscan;

	pscan = shm_toc_allocate(pcxt->toc, node->pscan_len);
	table_parallelscan_initialize(node->ss.ss_currentRelation,
								  pscan,
								  estate->es_snapshot);
	shm_toc_insert(pcxt->toc, node->ss.ps.plan->plan_node_id, pscan);
	node->ss.ss_currentScanDesc =
		table_beginscan_parallel(node->ss.ss_currentRelation, pscan);
}

/* ----------------------------------------------------------------
 *		ExecSeqScanReInitializeDSM
 *
 *		Reset shared state before beginning a fresh scan.
 * ----------------------------------------------------------------
 */
void
ExecSeqScanReInitializeDSM(SeqScanState *node,
						   ParallelContext *pcxt)
{
	ParallelTableScanDesc pscan;

	pscan = node->ss.ss_currentScanDesc->rs_parallel;
	table_parallelscan_reinitialize(node->ss.ss_currentRelation, pscan);
}

/* ----------------------------------------------------------------
 *		ExecSeqScanInitializeWorker
 *
 *		Copy relevant information from TOC into planstate.
 * ----------------------------------------------------------------
 */
void
ExecSeqScanInitializeWorker(SeqScanState *node,
							ParallelWorkerContext *pwcxt)
{
	ParallelTableScanDesc pscan;

	pscan = shm_toc_lookup(pwcxt->toc, node->ss.ps.plan->plan_node_id, false);
	node->ss.ss_currentScanDesc =
		table_beginscan_parallel(node->ss.ss_currentRelation, pscan);
}
